/* -*-Asm-*- */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999,2000,2001,2002,2005,2006,2007  Free Software Foundation, Inc.
 *  Copyright (C) 2008  Robert Millan
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "boot.h"
	
/*
 *  defines for the code go here
 */

	/* Absolute addresses
	   This makes the assembler generate the address without support
	   from the linker. (ELF can't relocate 16-bit addresses!) */
#define ABS(x) (x-_start+RELOC_ADDR)

	/* Print message string */
#define MSG(x)	movw $ABS(x), %si; call message

	.file	"boot.S"

	.text

	/* Tell GAS to generate 16-bit instructions so that this code works
	   in real mode. */
	.code16

.globl _start; _start:
	/*
	 * _start is loaded at 0x7c00 and is jumped to with CS:IP 0:0x7c00
	 */

	/*
	 * Beginning of the sector is compatible with the FAT/HPFS BIOS
	 * parameter block.
	 */

	jmp	after_BPB
	nop	/* do I care about this ??? */

	/*
	 * This space is for the BIOS parameter block!!!!  Don't change
	 * the first jump, nor start the code anywhere but right after
	 * this area.
	 */

	. = _start + 4

	/* scratch space */
mode:
	.byte	0
disk_address_packet:	
sectors:
	.long	0
heads:
	.long	0
cylinders:
	.word	0
sector_start:
	.byte	0
head_start:
	.byte	0
cylinder_start:
	.word	0
	/* more space... */

	. = _start + GRUB_BOOT_MACHINE_BPB_END

	/*
	 * End of BIOS parameter block.
	 */

kernel_address:
	.word	GRUB_BOOT_MACHINE_KERNEL_ADDR
kernel_segment:
	.word	GRUB_BOOT_MACHINE_KERNEL_SEG
kernel_sector:
	.long	0, 0
initial_dx:
	.word	0

after_BPB:

/* general setup */
	cli		/* we're not safe here! */

        /*
         * This is a workaround for buggy BIOSes which don't pass boot
         * drive correctly. If GRUB is installed into a HDD, check if
         * DL is masked correctly. If not, assume that the BIOS passed
         * a bogus value and set DL to 0x80, since this is the only
         * possible boot drive.
         */
        testb   $0x80, %dl
        jnz     1f
        movb    $0x80, %dl
1:
	
	/* set up %ds and %ss as offset from 0 */
	xorw	%ax, %ax
	movw	%ax, %ds
	movw	%ax, %ss

	/* Relocate ourselves.  */
	movw	$0x7c00, %si
	movw	$RELOC_ADDR, %di
	movw	$0x100, %cx
	cld
	rep
	movsw
	
	/*
	 * ljmp to the next instruction after relocation.  Note: we need
	 * this anyway because some bogus BIOSes jump to 07C0:0000 instead
	 * of 0000:7C00.
	 */
	ljmp	$0, $ABS(real_start)

real_start:	

	/* set up the REAL stack */
	movw	$GRUB_BOOT_MACHINE_STACK_SEG, %sp

	sti		/* we're safe again */

	/* save drive reference first thing! */
	movw	%dx, ABS(initial_dx)

	/* print a notification message on the screen */
	MSG(notification_string)

	/*
	 *  Now we prepare the environment read_sector() relies on.
	 */

	/* set %si to the disk address packet */
	movw	$ABS(disk_address_packet), %si

	/* check if LBA is supported */
	movb	$0x41, %ah
	movw	$0x55aa, %bx
	int	$0x13
	jc	hd_probe_error
	
	/* 
	 *  %dl may have been clobbered by INT 13, AH=41H.
	 *  This happens, for example, with AST BIOS 1.04.
	 */
	movw	ABS(initial_dx), %dx

	xorw	%ax, %ax
	/* the segment of buffer address */
	movw	%ax, 6(%si)
	movw	$0x200, 4(%si)

	incw	%ax	
	/* set the mode to non-zero */
	movb	%al, -1(%si)
	
	/* the blocks */
	movw	%ax, 2(%si)

	/* the size and the reserved byte */
	movw	$0x0010, (%si)

	xorl	%ebx, %ebx
	movl	%ebx, 12(%si)

	/* Read the GPT header.  */
	movl	$1, 8(%si)
	call	read_sector

	/* Check the header signature.  */
	cmpl	$0x20494645, 0x200	/* gpt_header.magic */
	jne	signature_error
	cmpl	$0x54524150, 0x204
	jne	signature_error

	/* Where's the first partition entry?  */
	movl	0x248, %ebx		/* gpt_header.partitions */
	movl	%ebx, 8(%si)
	call	read_sector

	xorl	%ecx, %ecx
	jmp	1f

next_partition:
	/*
	 * FIXME: Assuming gpt_header.partentry_size == 0x80
	 * (and gpt_header.maxpart == 4)
	 * It is easy to obtain the correct value, but if it were not
	 * sector-aligned, it can be a mess to access it (I don't think that'd
	 * ever fit).
	 */
	addl	$0x80, %ecx
1:

	/* If this is not a BIOS boot partition, jump to the next one.  */
	cmpl	$0x21686148, 0x200(%ecx)	/* gpt_partentry.type */
	jne	next_partition
	cmpl	$0x6e6f6449, 0x204(%ecx)
	jne	next_partition
	cmpl	$0x65654e74, 0x208(%ecx)
	jne	next_partition
	cmpl	$0x49464564, 0x20c(%ecx)
	jne	next_partition

	/* Finally, found something bootable!  */
	movl	0x220(%ecx), %ebx		/* gpt_partentry.start */
	movl	%ebx, 8(%si)

done:
	/* Now we read the actual payload.  First, set the offset to 0...  */
	xorw	%ax, %ax
	movw	%ax, 4(%si)
	/* ...and the segment to our boot area.  */
	movw	$GRUB_BOOT_MACHINE_BUFFER_SEG, 6(%si)
	call	read_sector

	movw	$GRUB_BOOT_MACHINE_BUFFER_SEG, %bx

	movw	ABS(kernel_segment), %es

	/*
	 * We need to save %cx and %si because the startup code in
	 * kernel uses them without initializing them.
	 */
	pusha
	pushw	%ds
	
	movw	$0x100, %cx
	movw	%bx, %ds
	xorw	%si, %si
	xorw	%di, %di
	
	cld
	
	rep
	movsw

	popw	%ds
	popa
	movw	ABS(initial_dx), %dx

	/* boot kernel */
	jmp	*(kernel_address)

/* END OF MAIN LOOP */

/*
 * BIOS call "INT 0x13 Function 0x42" to read sectors from disk into memory
 *	Call with	%ah = 0x42
 *			%dl = drive number
 *			%ds:%si = segment:offset of disk address packet
 *	Return:
 *			%al = 0x0 on success; err code on failure
 */

read_sector:
	movb	$0x42, %ah
	int	$0x13
	jc	hd_probe_error
	ret

/*
 * BIOS Geometry translation error (past the end of the disk geometry!).
 */
geometry_error:
	MSG(geometry_error_string)
	jmp	general_error

/*
 * Disk probe failure.
 */
hd_probe_error:
	MSG(hd_probe_error_string)
	jmp	general_error

/*
 * signature error
 */
signature_error:
	MSG(signature_error_string)
	jmp	general_error

/*
 * Read error on the disk.
 */
read_error:
	MSG(read_error_string)

general_error:
	MSG(general_error_string)

/* go here when you need to stop the machine hard after an error condition */
        /* tell the BIOS a boot failure, which may result in no effect */
        int	$0x18
stop:	jmp	stop

notification_string:	.string "GPT-MBR "
geometry_error_string:	.string "Geom"
hd_probe_error_string:	.string "Hdd"
signature_error_string:	.string "Sign"
read_error_string:	.string "Read"
general_error_string:	.string " Error"

/*
 * message: write the string pointed to by %si
 *
 *   WARNING: trashes %si, %ax, and %bx
 */

	/*
	 * Use BIOS "int 10H Function 0Eh" to write character in teletype mode
	 *	%ah = 0xe	%al = character
	 *	%bh = page	%bl = foreground color (graphics modes)
	 */
1:
	movw	$0x0001, %bx
	movb	$0xe, %ah
	int	$0x10		/* display a byte */
message:
	lodsb
	cmpb	$0, %al
	jne	1b	/* if not end of string, jmp to display */
	ret

	/*
	 *  Windows NT breaks compatibility by embedding a magic
	 *  number here.
	 */

	. = _start + GRUB_BOOT_MACHINE_WINDOWS_NT_MAGIC
nt_magic:	
	.long 0
	.word 0

part_start:	
	. = _start + GRUB_BOOT_MACHINE_PART_START

	/* Bootable flag.  */
	.byte 0x00

	/* CHS address of partition start.  */
	.byte 0x01, 0x01, 0x00

	/* Partition type.  */
	.byte 0xee

	/* CHS address of partition end.  */
	.byte 0xfe, 0xff, 0xff

	/* Beginning of the partition.  */
	.long 0x3f

	/* Length of the partition.  */
	.long 0xffffffff

	. = _start + GRUB_BOOT_MACHINE_PART_END

/* the last 2 bytes in the sector 0 contain the signature */
	.word	GRUB_BOOT_MACHINE_SIGNATURE
